@@ -9,6 +9,9 @@ module Network.WebSockets.Connection
     , defaultAcceptRequest
     , acceptRequestWith
     , rejectRequest
+    , RejectRequest(..)
+    , defaultRejectRequest
+    , rejectRequestWith
 
     , Connection (..)
 
@@ -30,28 +33,48 @@ module Network.WebSockets.Connection
     , sendPing
 
     , forkPingThread
+
+    , CompressionOptions (..)
+    , PermessageDeflate (..)
+    , defaultPermessageDeflate
+
+    , SizeLimit (..)
     ) where
 
 
 --------------------------------------------------------------------------------
-import qualified Blaze.ByteString.Builder    as Builder
-import           Control.Concurrent          (forkIO, threadDelay)
-import           Control.Exception           (AsyncException, fromException,
-                                              handle, throwIO)
-import           Control.Monad               (unless, when)
-import qualified Data.ByteString             as B
-import           Data.IORef                  (IORef, newIORef, readIORef,
-                                              writeIORef)
-import           Data.List                   (find)
-import qualified Data.Text                   as T
-import           Data.Word                   (Word16)
+import qualified Data.ByteString.Builder                         as Builder
+import           Control.Applicative                             ((<$>))
+import           Control.Concurrent                              (forkIO,
+                                                                  threadDelay)
+import           Control.Exception                               (AsyncException,
+                                                                  fromException,
+                                                                  handle,
+                                                                  throwIO)
+import           Control.Monad                                   (foldM, unless,
+                                                                  when)
+import qualified Data.ByteString                                 as B
+import qualified Data.ByteString.Char8                           as B8
+import           Data.IORef                                      (IORef,
+                                                                  newIORef,
+                                                                  readIORef,
+                                                                  writeIORef)
+import           Data.List                                       (find)
+import           Data.Maybe                                      (catMaybes)
+import qualified Data.Text                                       as T
+import           Data.Word                                       (Word16)
+import           Prelude
 
 
 --------------------------------------------------------------------------------
+import           Network.WebSockets.Connection.Options
+import           Network.WebSockets.Extensions                   as Extensions
+import           Network.WebSockets.Extensions.PermessageDeflate
+import           Network.WebSockets.Extensions.StrictUnicode
 import           Network.WebSockets.Http
 import           Network.WebSockets.Protocol
-import           Network.WebSockets.Stream   (Stream)
-import qualified Network.WebSockets.Stream   as Stream
+import           Network.WebSockets.Stream                       (Stream)
+import qualified Network.WebSockets.Stream                       as Stream
 import           Network.WebSockets.Types
 
 
@@ -81,7 +104,7 @@ data AcceptRequest = AcceptRequest
     -- ^ The subprotocol to speak with the client.  If 'pendingSubprotcols' is
     -- non-empty, 'acceptSubprotocol' must be one of the subprotocols from the
     -- list.
-    , acceptHeaders :: !Headers
+    , acceptHeaders     :: !Headers
     -- ^ Extra headers to send with the response.
     }
 
@@ -113,16 +136,48 @@ acceptRequestWith pc ar = case find (flip compatible request) protocols of
         sendResponse pc $ response400 versionHeader ""
         throwIO NotSupported
     Just protocol -> do
+
+        -- Get requested list of exceptions from client.
+        rqExts <- either throwIO return $
+            getRequestSecWebSocketExtensions request
+
+        -- Set up permessage-deflate extension if configured.
+        pmdExt <- case connectionCompressionOptions (pendingOptions pc) of
+            NoCompression                     -> return Nothing
+            PermessageDeflateCompression pmd0 ->
+                case negotiateDeflate (connectionMessageDataSizeLimit options) (Just pmd0) rqExts of
+                    Left err   -> do
+                        rejectRequestWith pc defaultRejectRequest {rejectMessage = B8.pack err}
+                        throwIO NotSupported
+                    Right pmd1 -> return (Just pmd1)
+
+        -- Set up strict utf8 extension if configured.
+        let unicodeExt =
+                if connectionStrictUnicode (pendingOptions pc)
+                    then Just strictUnicode else Nothing
+
+        -- Final extension list.
+        let exts = catMaybes [pmdExt, unicodeExt]
+
         let subproto = maybe [] (\p -> [("Sec-WebSocket-Protocol", p)]) $ acceptSubprotocol ar
-            headers = subproto ++ acceptHeaders ar
+            headers = subproto ++ acceptHeaders ar ++ concatMap extHeaders exts
             response = finishRequest protocol request headers
-        sendResponse pc response
-        parse <- decodeMessages protocol (pendingStream pc)
-        write <- encodeMessages protocol ServerConnection (pendingStream pc)
+
+        either throwIO (sendResponse pc) response
+
+        parseRaw <- decodeMessages
+            protocol
+            (connectionFramePayloadSizeLimit options)
+            (connectionMessageDataSizeLimit options)
+            (pendingStream pc)
+        writeRaw <- encodeMessages protocol ServerConnection (pendingStream pc)
+
+        write <- foldM (\x ext -> extWrite ext x) writeRaw exts
+        parse <- foldM (\x ext -> extParse ext x) parseRaw exts
 
         sentRef    <- newIORef False
         let connection = Connection
-                { connectionOptions   = pendingOptions pc
+                { connectionOptions   = options
                 , connectionType      = ServerConnection
                 , connectionProtocol  = protocol
                 , connectionParse     = parse
@@ -133,14 +188,59 @@ acceptRequestWith pc ar = case find (flip compatible request) protocols of
         pendingOnAccept pc connection
         return connection
   where
+    options       = pendingOptions pc
     request       = pendingRequest pc
     versionHeader = [("Sec-WebSocket-Version",
         B.intercalate ", " $ concatMap headerVersions protocols)]
 
 
 --------------------------------------------------------------------------------
-rejectRequest :: PendingConnection -> B.ByteString -> IO ()
-rejectRequest pc message = sendResponse pc $ response400 [] message
+-- | Parameters that allow you to tweak how a request is rejected.  Please use
+-- 'defaultRejectRequest' and modify fields using record syntax so your code
+-- will not break when new fields are added.
+data RejectRequest = RejectRequest
+    { -- | The status code, 400 by default.
+      rejectCode    :: !Int
+    , -- | The message, "Bad Request" by default
+      rejectMessage :: !B.ByteString
+    , -- | Extra headers to be sent with the response.
+      rejectHeaders :: Headers
+    , -- | Reponse body of the rejection.
+      rejectBody    :: !B.ByteString
+    }
+
+
+--------------------------------------------------------------------------------
+defaultRejectRequest :: RejectRequest
+defaultRejectRequest = RejectRequest
+    { rejectCode    = 400
+    , rejectMessage = "Bad Request"
+    , rejectHeaders = []
+    , rejectBody    = ""
+    }
+
+
+--------------------------------------------------------------------------------
+rejectRequestWith
+    :: PendingConnection  -- ^ Connection to reject
+    -> RejectRequest      -- ^ Params on how to reject the request
+    -> IO ()
+rejectRequestWith pc reject = sendResponse pc $ Response
+    ResponseHead
+        { responseCode    = rejectCode reject
+        , responseMessage = rejectMessage reject
+        , responseHeaders = rejectHeaders reject
+        }
+    (rejectBody reject)
+
+
+--------------------------------------------------------------------------------
+rejectRequest
+    :: PendingConnection  -- ^ Connection to reject
+    -> B.ByteString       -- ^ Rejection response body
+    -> IO ()
+rejectRequest pc body = rejectRequestWith pc
+    defaultRejectRequest {rejectBody = body}
 
 
 --------------------------------------------------------------------------------
@@ -159,22 +259,6 @@ data Connection = Connection
     }
 
 
---------------------------------------------------------------------------------
--- | Set options for a 'Connection'.
-data ConnectionOptions = ConnectionOptions
-    { connectionOnPong :: !(IO ())
-      -- ^ Whenever a 'pong' is received, this IO action is executed. It can be
-      -- used to tickle connections or fire missiles.
-    }
-
-
---------------------------------------------------------------------------------
-defaultConnectionOptions :: ConnectionOptions
-defaultConnectionOptions = ConnectionOptions
-    { connectionOnPong = return ()
-    }
-
-
 --------------------------------------------------------------------------------
 receive :: Connection -> IO Message
 receive conn = do
@@ -199,8 +283,8 @@ receiveDataMessage :: Connection -> IO DataMessage
 receiveDataMessage conn = do
     msg <- receive conn
     case msg of
-        DataMessage am    -> return am
-        ControlMessage cm -> case cm of
+        DataMessage _ _ _ am -> return am
+        ControlMessage cm    -> case cm of
             Close i closeMsg -> do
                 hasSentClose <- readIORef $ connectionSentClose conn
                 unless hasSentClose $ send conn msg
@@ -216,11 +300,7 @@ receiveDataMessage conn = do
 --------------------------------------------------------------------------------
 -- | Receive a message, converting it to whatever format is needed.
 receiveData :: WebSocketsData a => Connection -> IO a
-receiveData conn = do
-    dm <- receiveDataMessage conn
-    case dm of
-        Text x   -> return (fromLazyByteString x)
-        Binary x -> return (fromLazyByteString x)
+receiveData conn = fromDataMessage <$> receiveDataMessage conn
 
 
 --------------------------------------------------------------------------------
@@ -229,6 +309,7 @@ send conn = sendAll conn . return
 
 --------------------------------------------------------------------------------
 sendAll :: Connection -> [Message] -> IO ()
+sendAll _    []   = return ()
 sendAll conn msgs = do
     when (any isCloseMessage msgs) $
       writeIORef (connectionSentClose conn) True
@@ -238,32 +319,41 @@ sendAll conn msgs = do
     isCloseMessage _                            = False
 
 --------------------------------------------------------------------------------
--- | Send a 'DataMessage'
+-- | Send a 'DataMessage'.  This allows you send both human-readable text and
+-- binary data.  This is a slightly more low-level interface than 'sendTextData'
+-- or 'sendBinaryData'.
 sendDataMessage :: Connection -> DataMessage -> IO ()
 sendDataMessage conn = sendDataMessages conn . return
 
 --------------------------------------------------------------------------------
--- | Send a collection of 'DataMessage's
+-- | Send a collection of 'DataMessage's.  This is more efficient than calling
+-- 'sendDataMessage' many times.
 sendDataMessages :: Connection -> [DataMessage] -> IO ()
-sendDataMessages conn = sendAll conn . map DataMessage
+sendDataMessages conn = sendAll conn . map (DataMessage False False False)
 
 --------------------------------------------------------------------------------
--- | Send a message as text
+-- | Send a textual message.  The message will be encoded as UTF-8.  This should
+-- be the default choice for human-readable text-based protocols such as JSON.
 sendTextData :: WebSocketsData a => Connection -> a -> IO ()
 sendTextData conn = sendTextDatas conn . return
 
 --------------------------------------------------------------------------------
--- | Send a collection of messages as text
+-- | Send a number of textual messages.  This is more efficient than calling
+-- 'sendTextData' many times.
 sendTextDatas :: WebSocketsData a => Connection -> [a] -> IO ()
-sendTextDatas conn = sendDataMessages conn . map (Text . toLazyByteString)
+sendTextDatas conn =
+    sendDataMessages conn .
+    map (\x -> Text (toLazyByteString x) Nothing)
 
 --------------------------------------------------------------------------------
--- | Send a message as binary data
+-- | Send a binary message.  This is useful for sending binary blobs, e.g.
+-- images, data encoded with MessagePack, images...
 sendBinaryData :: WebSocketsData a => Connection -> a -> IO ()
 sendBinaryData conn = sendBinaryDatas conn . return
 
 --------------------------------------------------------------------------------
--- | Send a collection of messages as binary data
+-- | Send a number of binary messages.  This is more efficient than calling
+-- 'sendBinaryData' many times.
 sendBinaryDatas :: WebSocketsData a => Connection -> [a] -> IO ()
 sendBinaryDatas conn = sendDataMessages conn . map (Binary . toLazyByteString)
 
@@ -299,6 +389,10 @@ sendPing conn = send conn . ControlMessage . Ping . toLazyByteString
 --------------------------------------------------------------------------------
 -- | Forks a ping thread, sending a ping message every @n@ seconds over the
 -- connection. The thread dies silently if the connection crashes or is closed.
+--
+-- This is useful to keep idle connections open through proxies and whatnot.
+-- Many (but not all) proxies have a 60 second default timeout, so based on that
+-- sending a ping every 30 seconds is a good idea.
 forkPingThread :: Connection -> Int -> IO ()
 forkPingThread conn n
     | n <= 0    = return ()
