@@ -7,6 +7,7 @@ module Network.WebSockets.Stream
     , makeSocketStream
     , makeEchoStream
     , parse
+    , parseBin
     , write
     , close
     ) where
@@ -14,11 +15,12 @@ module Network.WebSockets.Stream
 import           Control.Concurrent.MVar        (MVar, newEmptyMVar, newMVar,
                                                  putMVar, takeMVar, withMVar)
 import           Control.Exception              (onException, throwIO)
-import           Control.Monad                  (forM_, when)
+import           Control.Monad                  (forM_)
 import qualified Data.Attoparsec.ByteString     as Atto
+import qualified Data.Binary.Get                as BIN
 import qualified Data.ByteString                as B
 import qualified Data.ByteString.Lazy           as BL
-import           Data.IORef                     (IORef, atomicModifyIORef,
+import           Data.IORef                     (IORef, atomicModifyIORef',
                                                  newIORef, readIORef,
                                                  writeIORef)
 import qualified Network.Socket                 as S
@@ -73,27 +75,31 @@ makeStream receive send = do
     return $ Stream (receive' ref receiveLock) (send' ref sendLock) ref
   where
     closeRef :: IORef StreamState -> IO ()
-    closeRef ref = atomicModifyIORef ref $ \state -> case state of
+    closeRef ref = atomicModifyIORef' ref $ \state -> case state of
         Open   buf -> (Closed buf, ())
         Closed buf -> (Closed buf, ())
 
-    assertNotClosed :: IORef StreamState -> IO a -> IO a
-    assertNotClosed ref io = do
+    -- Throw a 'ConnectionClosed' is the connection is not 'Open'.
+    assertOpen :: IORef StreamState -> IO ()
+    assertOpen ref = do
         state <- readIORef ref
         case state of
             Closed _ -> throwIO ConnectionClosed
-            Open   _ -> io
+            Open   _ -> return ()
 
     receive' :: IORef StreamState -> MVar () -> IO (Maybe B.ByteString)
-    receive' ref lock = withMVar lock $ \() -> assertNotClosed ref $ do
+    receive' ref lock = withMVar lock $ \() -> do
+        assertOpen ref
         mbBs <- onException receive (closeRef ref)
         case mbBs of
             Nothing -> closeRef ref >> return Nothing
             Just bs -> return (Just bs)
 
     send' :: IORef StreamState -> MVar () -> (Maybe BL.ByteString -> IO ())
-    send' ref lock mbBs = withMVar lock $ \() -> assertNotClosed ref $ do
-        when (mbBs == Nothing) (closeRef ref)
+    send' ref lock mbBs = withMVar lock $ \() -> do
+        case mbBs of
+            Nothing -> closeRef ref
+            Just _  -> assertOpen ref
         onException (send mbBs) (closeRef ref)
 
 
@@ -102,7 +108,7 @@ makeSocketStream :: S.Socket -> IO Stream
 makeSocketStream socket = makeStream receive send
   where
     receive = do
-        bs <- SB.recv socket 1024
+        bs <- SB.recv socket 8192
         return $ if B.null bs then Nothing else Just bs
 
     send Nothing   = return ()
@@ -124,6 +130,38 @@ makeEchoStream = do
 
 
 --------------------------------------------------------------------------------
+parseBin :: Stream -> BIN.Get a -> IO (Maybe a)
+parseBin stream parser = do
+    state <- readIORef (streamState stream)
+    case state of
+        Closed remainder
+            | B.null remainder -> return Nothing
+            | otherwise        -> go (BIN.runGetIncremental parser `BIN.pushChunk` remainder) True
+        Open buffer
+            | B.null buffer -> do
+                mbBs <- streamIn stream
+                case mbBs of
+                    Nothing -> do
+                        writeIORef (streamState stream) (Closed B.empty)
+                        return Nothing
+                    Just bs -> go (BIN.runGetIncremental parser `BIN.pushChunk` bs) False
+            | otherwise     -> go (BIN.runGetIncremental parser `BIN.pushChunk` buffer) False
+  where
+    -- Buffer is empty when entering this function.
+    go (BIN.Done remainder _ x) closed = do
+        writeIORef (streamState stream) $
+            if closed then Closed remainder else Open remainder
+        return (Just x)
+    go (BIN.Partial f) closed
+        | closed    = go (f Nothing) True
+        | otherwise = do
+            mbBs <- streamIn stream
+            case mbBs of
+                Nothing -> go (f Nothing) True
+                Just bs -> go (f (Just bs)) False
+    go (BIN.Fail _ _ err) _ = throwIO (ParseException err)
+
+
 parse :: Stream -> Atto.Parser a -> IO (Maybe a)
 parse stream parser = do
     state <- readIORef (streamState stream)
